package msigi.netclient;

import java.io.*;
import java.net.*;

import msigi.netclient.*;

public class NetworkClientThreads implements Runnable {

	// Variables
	private Thread netThreadRunner = null;
	private Thread receiveThreadRunner = null;
	private Object caller = null;
	private NetworkClientState st;
	private Socket clientSocket = null;
	private boolean operationInterrupted = false;
	private String outputData;
	private PrintWriter out;
	private ReceiveData receiveData = null;

	// Constructor
	public NetworkClientThreads(Object callerParam, NetworkClientState stParam) {
		caller = callerParam;
		st = stParam;
	}

	// Starts a thread that connects to the server
	public synchronized void connect() throws TrivialCommunicationException {
		if (clientSocket == null) {
			netThreadRunner = new Thread(this, "NetworkThread");
			netThreadRunner.start();
		} else {
			TrivialCommunicationException ex;
			ex = new TrivialCommunicationException("Attempt to reconnect an already connected object");
			throw (ex);
		}
	}

	// Main execution function which establishes a connection to the server
	public synchronized void run() {
		try {
			InetSocketAddress sockAddr = new InetSocketAddress(st.remoteHost, st.remotePort);
			clientSocket = new Socket();
			clientSocket.connect(sockAddr, 15000); // timeout is 15 seconds
			initializeConnection();
			waitForOutputData();
		} catch (UnknownHostException ex) {
			st.errorMessage = "Unknown host: " + ex.getMessage();
		} catch (IOException ex) {
			st.errorMessage = "IOException while trying to connect: " + ex.getMessage();
		} finally {
			//System.out.println(gtn() + ": in finally branch");
			if(clientSocket != null) {
				try {
			//System.out.println(gtn() + ": trying clientSocket.close()");
					clientSocket.close();
			//System.out.println(gtn() + ": success");
				} catch (IOException ex) {
					// do nothing
				} finally {
					clientSocket = null;
			//System.out.println(gtn() + ": clientSocket is now null");
				}
			}
			//System.out.println(gtn() + ": notifying caller");
			if (st.isAttemptingToConnect) {
				synchronized (caller) { caller.notify(); }
			}
			//System.out.println(gtn() + ": done notifying caller");
		}
	}

	// Set up the input and output streams for the network connection and read welcome string
	private synchronized void initializeConnection() {
		try {
			out = new PrintWriter(clientSocket.getOutputStream(), true);
			st.isConnected = true;
			receiveData = new ReceiveData();
			receiveThreadRunner = new Thread(receiveData,"ReceiveThread");
			receiveThreadRunner.start();

			// read the welcome line sent by the server
			st.serverVersionString = receiveData.getLine(); 

		} catch (FatalCommunicationException ex) {
			st.errorMessage = ex.getMessage();
		} catch (IOException ex) {
			st.errorMessage = "IOException in inializeConnection(): " + ex.getMessage();
		} finally {
			synchronized (caller) {
				caller.notify();
			}
		}
	}

	// Wait for output and then send it to the server
	private synchronized void waitForOutputData() {
		try {
			while (st.isConnected) {
		//System.out.println(gtn() + ": entered while (st.isConnected), about to wait()");
				wait();
		//System.out.println(gtn() + ": returned from wait(), st.isConnected=" + st.isConnected);
				if (st.isConnected) {
					out.println(outputData);
				}
			}
		} catch (InterruptedException ex) {
			// Ignore the exception since it should never occur anyway
			//System.out.println("InterruptedException in waitForOutputData()"); 
		}
	}

	// Sends a line over the network
	public synchronized void sendLine(String line) {
		outputData = line;
		notify();
	}

	// Receives a line from the network (this must NOT be declared synchronized, otherwise
	// data waiting to be sent will be delayed as long as this function blocks, possibly
	// resulting in a deadlock)
	public String getLine() throws FatalCommunicationException {
		return receiveData.getLine();
	}

	// Closes the network connection
	public synchronized void closeConnection() {
		// FIX: sometimes control does not enter this routine (deadlock)
		//System.out.println(gtn() + ": entered closeConnection(), st.isConnected=" + st.isConnected);
		st.isConnected = false;
		notifyAll();
	}

	// Class to run as a thread for receiving data over the network
	public class ReceiveData implements Runnable {

		// Variables
		private BufferedReader in = null;
		private String buf = null;

		// Main executing function
		//public synchronized void run() {
		public void run() {
			String buftmp;
			try {
				in = new BufferedReader(new InputStreamReader(clientSocket.getInputStream()));
				while(st.isConnected) {
					buftmp = in.readLine();
					synchronized (this) {
						buf = buftmp;
						if(buf == null) {
							// Connection closed by peer
							closeConnection();
							notify();
						} else {
							// A line was successfully received
							notify();
							try {
								wait();
							} catch (InterruptedException ex) {
							}
						}
					}
				}
			} catch (IOException ex) {
				// Error on connection
				synchronized (this) {
					closeConnection();
					notify();
				}
			}
		}

		// Function for reading a line from the network
		private synchronized String getLine() throws FatalCommunicationException {
			String myBuffer;
			while (buf == null && st.isConnected) {
				// Go to sleep until a string is present
				try {
					wait(30000);
					if (buf == null && st.isConnected) {
						// Operation timed out
						closeConnection();
						throw (new FatalCommunicationException(
								       "Receive operation timed out"));	
					}
				} catch (InterruptedException ex) {
				}
			}
			if (st.isConnected) {
				myBuffer = buf;
				buf = null;
				notify();
				return myBuffer;
			} else {
				notify();
				FatalCommunicationException ex =
					new FatalCommunicationException("Connection was closed by peer");
				throw (ex);
			}
		}
	}

	// For debugging only: returns the thread name as a string
	private String gtn() {
		Thread blah = Thread.currentThread();
		return blah.getName();
	}

}
