package msigi.netgpibc;

import msigi.netgpibc.*;
import msigi.netclient.*;

public class GPIBCommunicationThread implements Runnable {

	// Length of command queue
	private static final int QUEUE_LENGTH = 8;

	// Error constants
	private static final int CMD_NONE = 0;
	private static final int CMD_CONNECT = 1;
	private static final int CMD_CLOSE = 2;
	private static final int CMD_WRITE = 3;
	private static final int CMD_READ = 4;
	private static final int CMD_QUERY = 5;

	// Command queue
	private int[] commands;
	private String[] params;
	private CommStatusHandler[] handlers;
	private int qhead = 0;
	private int qtail = 0;

	// GPIB communication related variables
	private String remoteHost = null;
	private int remotePort = 0;
	private int gpibAddress = 0;
	private NetGPIBClient gpib = null;

	// Thread related variables
	private Thread gpibThreadRunner = null; 

	// Constructor
	public GPIBCommunicationThread() {
		commands = new int[QUEUE_LENGTH];
		params = new String[QUEUE_LENGTH];
		handlers = new CommStatusHandler[QUEUE_LENGTH];
		gpibThreadRunner = new Thread(this, "GPIBThread");
		gpibThreadRunner.start();
	}

	// Main executing function for thread
	public void run() {
		boolean hasCommandFlag = false;
		try {
			while (true) {
				// Check if there was a command, if so, take action and restart loop
				synchronized (this) {
 					if (qhead == qtail) {
						// Wait for a command
						wait();
						hasCommandFlag = false;
					} else {
						// There was a command, set a flag to make it execute
						hasCommandFlag = true;
					}

				}	
				if (hasCommandFlag) {
					// There was a command, execute it
					executeCommand(); 
				}

			}
		} catch (InterruptedException ex) {
		}
	}

	// Actual code to do command execution
	public void executeCommand() {
		CommStatus status = new CommStatus();
		synchronized (this) {
			increment_qhead(); 
			//System.out.println("executing command " + qhead);
		}
		if (commands[qhead] == CMD_CONNECT) {
			try {
				gpib = new NetGPIBClient(remoteHost, remotePort, gpibAddress);
				gpib.connect();
				status.isConnected = true;
			} catch (FatalCommunicationException ex) {
				status.isConnected = false;
				status.errorNum = CommStatus.ERROR_COMM_FATAL;
				status.errorMessage = "Fatal communications error: " + ex.getMessage();
			} catch (TrivialCommunicationException ex) {
				status.isConnected = true;
				status.errorNum = CommStatus.ERROR_COMM_TRIVIAL;
				status.errorMessage = "Trivial communications error: " + ex.getMessage();
			} catch (GPIBException ex) {
				status.isConnected = false;
				status.errorNum = CommStatus.ERROR_GPIB;
				status.errorMessage = "GPIB error: " + ex.getMessage();
			}
			handlers[qhead].commStatusHandler(status);
		} else if (commands[qhead] == CMD_CLOSE) {
			try {
				//System.out.println("calling gpib.close()");
				gpib.close();
				status.isConnected = false;
				//System.out.println("returned from gpib.close()");
			} catch (FatalCommunicationException ex) {
				status.isConnected = false;
				status.errorNum = CommStatus.ERROR_COMM_FATAL;
				status.errorMessage = "Fatal communications error: " + ex.getMessage();
			}
			handlers[qhead].commStatusHandler(status);
		} else if (commands[qhead] == CMD_WRITE) {
			try {
				gpib.write(params[qhead]);
				status.isConnected = true;
			} catch (FatalCommunicationException ex) {
				status.isConnected = false;
				status.errorNum = CommStatus.ERROR_COMM_FATAL;
				status.errorMessage = "Fatal communications error: " + ex.getMessage();
			} catch (GPIBException ex) {
				status.isConnected = true;
				status.errorNum = CommStatus.ERROR_GPIB;
				status.errorMessage = "GPIB error: " + ex.getMessage();
			}
			handlers[qhead].commStatusHandler(status);
		} else if (commands[qhead] == CMD_READ) {
			try {
				status.readMessage = gpib.read();
				status.isConnected = true;
			} catch (FatalCommunicationException ex) {
				status.isConnected = false;
				status.errorNum = CommStatus.ERROR_COMM_FATAL;
				status.errorMessage = "Fatal communications error: " + ex.getMessage();
			} catch (GPIBException ex) {
				status.isConnected = true;
				status.errorNum = CommStatus.ERROR_GPIB;
				status.errorMessage = "GPIB error: " + ex.getMessage();
			}
			handlers[qhead].commStatusHandler(status);
		} else if (commands[qhead] == CMD_QUERY) {
			try {
				status.readMessage = gpib.query(params[qhead]);
				status.isConnected = true;
			} catch (FatalCommunicationException ex) {
				status.isConnected = false;
				status.errorNum = CommStatus.ERROR_COMM_FATAL;
				status.errorMessage = "Fatal communications error: " + ex.getMessage();
			} catch (GPIBException ex) {
				status.isConnected = true;
				status.errorNum = CommStatus.ERROR_GPIB;
				status.errorMessage = "GPIB error: " + ex.getMessage();
			}
			handlers[qhead].commStatusHandler(status);
		}
	}

	// Store hostname, port, and gpib address for later use
	public synchronized void configure(String hostname, int port, int addr) {
		remoteHost = hostname;
		remotePort = port;
		gpibAddress = addr;
	}

	// Add connect command to queue
	public synchronized void connect(CommStatusHandler callback) throws TrivialCommunicationException {
		increment_qtail();
		commands[qtail] = CMD_CONNECT;
		handlers[qtail] = callback;
		//System.out.println("inserted connect() into position " + qtail);
		notify();
	}

	// Add close command to queue
	public synchronized void close(CommStatusHandler callback) throws TrivialCommunicationException {
		increment_qtail();
		commands[qtail] = CMD_CLOSE;
		handlers[qtail] = callback;
		//System.out.println("inserted close() into position " + qtail);
		notify();
	}

	// Add write command to queue
	public synchronized void write(CommStatusHandler callback, String cmd) throws TrivialCommunicationException {
		increment_qtail();
		commands[qtail] = CMD_WRITE;
		handlers[qtail] = callback;
		params[qtail] = cmd;
		notify();
	}

	// Add read command to queue
	public synchronized void read(CommStatusHandler callback) throws TrivialCommunicationException {
		increment_qtail();
		commands[qtail] = CMD_READ;
		handlers[qtail] = callback;
		//System.out.println("inserted read() into position " + qtail);
		notify();
	}

	// Add query command to queue
	public synchronized void query(CommStatusHandler callback, String cmd) throws TrivialCommunicationException {
		increment_qtail();
		commands[qtail] = CMD_QUERY;
		handlers[qtail] = callback;
		params[qtail] = cmd;
		notify();
	}

	// Increment index of the head of the queue
	private synchronized void increment_qhead() {
		qhead++;
		if (qhead == QUEUE_LENGTH) {
			qhead = 0;
		}
	}

	// Increment index of the tail of the queue
	private synchronized void increment_qtail() throws TrivialCommunicationException {
		int qtail2;
		qtail2 = qtail + 1;
		if (qtail2 == QUEUE_LENGTH) {
			qtail2 = 0;
		}
		while (qtail2 == qhead) {
			// block until qhead changes -- or throw exception?
			//System.out.println("out of buffer space"); 
			throw (new TrivialCommunicationException("Out of command queue space"));
		}
		qtail = qtail2;
	}
}
