package msigi.netgpibc;

import msigi.netgpibc.*;
import msigi.netclient.*;

public class NetGPIBClient {

	// Variables
	int gpibAddress;
	NetworkClientThreads netThread;
	NetworkClientState st;

	// Constructor
	public NetGPIBClient(String hostname, int port, int addr) {
		//System.out.println("constructing");
		gpibAddress = addr;
		st = new NetworkClientState(hostname, port);
		netThread = new NetworkClientThreads(this, st);
	}

	// Open a connection
	public synchronized void connect() throws FatalCommunicationException, 
			TrivialCommunicationException, GPIBException {
		if (st.isAttemptingToConnect) {
			TrivialCommunicationException ex;
			ex = new TrivialCommunicationException("Attempt to reconnect while a connect is in progress");
			throw (ex);
		} else {

			//System.out.println("connect()");
			st.errorMessage = null;
			st.isAttemptingToConnect = true;
			netThread.connect();

			// wait to be notified of connection success or failure
			try {
				wait(); 
			} catch (InterruptedException ex) {
				// FIX: check if connection needs to be closed?
				st.errorMessage = ex.getMessage();
			}
			st.isAttemptingToConnect = false;
			if (st.errorMessage != null) {
				// An error occurred
				throw (new FatalCommunicationException(st.errorMessage));
			}
			try {
				String response;
				netThread.sendLine("init " + gpibAddress);
				response = netThread.getLine();
				//System.out.println("Response: " + response);
				checkResponse(response);
			} catch (FatalCommunicationException ex) {
				netThread.closeConnection();
				throw (ex);
			} catch (GPIBException ex) {
				netThread.closeConnection();
				throw (ex);
			}
		} 
	}

	// Close connection
	public synchronized void close() throws FatalCommunicationException {
		//System.out.println("close()");
		if (st.isConnected) {
			//System.out.println(gtn() + ": sending close command");
			netThread.sendLine("close");
			//System.out.println(gtn() + ": getting reply to close command");
			String response = netThread.getLine();
			//System.out.println(gtn() + ": got reply to close request: " + response);
			//try { Thread.sleep(500); } catch (Exception ex) { }
			netThread.closeConnection();
			//System.out.println(gtn() + ": returned from netThread.closeConnection()");
		} else {
			FatalCommunicationException ex;
			ex = new FatalCommunicationException("Attempt to close unconnected GPIB object");
			throw (ex);
		}
	}

	// Close connection without waiting for server acknowlegement
	public synchronized void forceClose() {
		//netThread.sendLine("close"); // does not work
		netThread.closeConnection();
	}

	// Send command to instrument
	public synchronized void write(String cmd) throws FatalCommunicationException, GPIBException {
		if (st.isConnected) {
			//System.out.println("write(\"" + cmd + "\")");
			String response;
			netThread.sendLine("write " + cmd);
			response = netThread.getLine();
			//System.out.println("Response: " + response);
			checkResponse(response);
		} else {
			FatalCommunicationException ex;
			ex= new FatalCommunicationException("Attempt to write on unconnected GPIB object");
			throw (ex);
		}
	}

	// Read from instrument
	public synchronized String read() throws FatalCommunicationException, GPIBException {
		//System.out.println("read()");
		if (st.isConnected) {
			String response;
			String readstr;
			netThread.sendLine("read");
			response = netThread.getLine();
			readstr = parseReadResponse(response);
			return readstr;
		} else {
			FatalCommunicationException ex;
			ex = new FatalCommunicationException("Attempt to read on unconnected GPIB object");
			throw (ex);
		}
	}

	// Query instrument (send command and then read)
	public synchronized String query(String cmd) throws FatalCommunicationException, GPIBException {
		//System.out.println("query(\"" + cmd + "\")");
		if (st.isConnected) {
			String response;
			String readstr;
			netThread.sendLine("query " + cmd);
			response = netThread.getLine();
			readstr = parseReadResponse(response);
			return readstr;
		} else {
			FatalCommunicationException ex;
			ex = new FatalCommunicationException("Attempt to query on unconnected GPIB object");
			throw (ex);
		}
	}	

	// Check if the last command succeeded
	// Throws a GPIBException on failure
	// Upon protocol violation, closes the connections and throws a FatalCommunicationException 
	private void checkResponse (String resp) throws FatalCommunicationException,
			GPIBCommException, GPIBFatalException, GPIBLockException {
		String[] tokens = resp.split("\\s", 3);
		if (tokens.length < 1) {
			forceClose();
			throw (new FatalCommunicationException("Invalid response: no tokens"));
		}
		if (tokens[0].equals("ok")) {
			return;
		} else if (tokens[0].equals("error")) {
			if (tokens.length == 2) {
				throwErrorException(tokens[1]);
				//ex = new GPIBException("Error " + tokens[1] + " occurred");
			} else if (tokens.length == 3) {
				throwErrorException(tokens[1]);
				//ex = new GPIBException("Error " + tokens[1] + " occurred: " + tokens[2]);
			} else {
				forceClose();
				throw (new FatalCommunicationException("Unknown error"));
			}
		} else {
			forceClose();
			throw (new FatalCommunicationException("Invalid response: violates protocol"));
		}
	}

	// Parse a response to a read or query command
	// Returns the String received from the instrument or throws a GPIBException
	// Upon protocol violation, closes the connections and throws a FatalCommunicationException 
	private String parseReadResponse (String resp) throws FatalCommunicationException,
			GPIBCommException, GPIBFatalException, GPIBLockException {
		String[] tokens = resp.split("\\s", 2);
		if (tokens.length < 2) {
			forceClose();
			throw (new FatalCommunicationException("Invalid response: violates protocol"));
		}
		if (tokens[0].equals("read")) {
			return tokens[1];
		} else if (tokens[0].equals("error")) {
			String[] errinfo = tokens[1].split("\\s", 2);
			if (errinfo.length == 2) {
				throwErrorException(errinfo[0]);
				//ex = new GPIBException("Error " + errinfo[0] + " occurred: " + errinfo[1]);
			} else {
				throwErrorException(tokens[1]);
				//ex = new GPIBException("Error " + tokens[1] + " occurred");
			}
			//throw (ex);
		} else {
			forceClose();
			throw (new FatalCommunicationException("Invalid response: violates protocol"));
		}
		return "";
	}

	// An exception based on an error number received
	private void throwErrorException(String errnumstr) throws FatalCommunicationException,
			GPIBCommException, GPIBFatalException, GPIBLockException {
		int errnum = 0;
		try {
			errnum = Integer.parseInt(errnumstr);
		} catch (NumberFormatException ex) {
			forceClose();
			throw (new FatalCommunicationException("Invalid response: violates protocol"));
		}
		if (errnum == 31) {
			// Bad command
			forceClose();
			throw (new FatalCommunicationException("Client violated protocol"));
		} else if (errnum == 32) {
			// GPIB communications error
			throw (new GPIBCommException("GPIB communications error")); 
		} else if (errnum == 33) {
			// Fatal GPIB error
			throw (new GPIBFatalException("GPIB fatal error")); 
		} else if (errnum == 34) {
			// Cannot obtain instrument lock
			throw (new GPIBLockException("Instrument locked by another process")); 
		} else if (errnum == 35) {
			// Server had an internal error
			throw (new GPIBFatalException("GPIB server internal error")); 
		} else {
			// The server produced an unknown error
			forceClose();
			throw (new FatalCommunicationException("Unknown error on server"));
		}
	}	

	// Finalize method (this appears not to be run when program exits)
	//protected void finalize() {
	//	System.out.println("finalizing");
	//	//super.finalize();
	//}

	private String gtn() {
		Thread blah = Thread.currentThread();
		return blah.getName();
	}

}
